/*
  File: fs/extattr.c

  Extended attribute handling.
 
  Copyright (C) 2001 by Andreas Gruenbacher <a.gruenbacher@computer.org>
  Copyright (C) 2001 SGI - Silicon Graphics, Inc <linux-xfs@oss.sgi.com>
 */

#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/smp_lock.h>
#include <linux/extattr.h>

static long
extattr_inode(struct inode *i, int cmd, char *name, void *value, size_t size)
{
	int op, flags, error = -ENOTSUP;

	flags = (cmd >> EASHIFT) | EA_FLAG_USER;
	op = (cmd & EAMASK);

	lock_kernel();
	if (i->i_op && i->i_op->extattr)
		error = i->i_op->extattr(i, op, name, value, size, flags);
	unlock_kernel();
	return error;
}

asmlinkage long
sys_extattr(char *path, int cmd, char *name, void *value, size_t size)
{
	struct nameidata nd;
	int error;

	error = user_path_walk(path, &nd);
	if (error)
		return error;
	error = extattr_inode(nd.dentry->d_inode, cmd, name, value, size);
	path_release(&nd);
	return error;
}

asmlinkage long
sys_lextattr(char *path, int cmd, char *name, void *value, size_t size)
{
	struct nameidata nd;
	int error;

	error = user_path_walk_link(path, &nd);
	if (error)
		return error;
	error = extattr_inode(nd.dentry->d_inode, cmd, name, value, size);
	path_release(&nd);
	return error;
}

asmlinkage long
sys_fextattr(int fd, int cmd, char *name, void *value, size_t size)
{
	struct file *f;
	int error = -EBADF;

	f = fget(fd);
	if (!f)
		return error;
	error = extattr_inode(f->f_dentry->d_inode, cmd, name, value, size);
	fput(f);
	return error;
}
