/*
 * NOTE: This file is a bit hodge-podged - it is based on a much
 * cleaner version of quotactl(2) written by Jan Kara and myself
 * for his new VFS quota implementation (see Alan Cox's patches).
 * 
 * The quota interfaces in Linus' tree are not quite the same as
 * Jan's new ones, so I ended up with a bit of a gross hack here
 * - in time this will become less gross, once the new VFS quota
 * implementation makes its way into Linus' source tarball.
 * 
 * 	-- nathans@sgi.com
 */

/*
 * Quota code necessary even when VFS quota support is not compiled
 * into the kernel.  The interesting stuff is over in dquot.c, here
 * we have symbols for initial quotactl(2) handling, the sysctl(2)
 * variables, etc - things needed even when quota support disabled.
 */

#include <linux/fs.h>
#include <asm/current.h>
#include <asm/uaccess.h>
#include <linux/kernel.h>
#include <linux/smp_lock.h>

int nr_dquots, nr_free_dquots;

#ifdef CONFIG_QUOTA	/* This is the hack letting us get into VFS quota */
extern int vfsqops(struct super_block *, kdev_t, int, int, int, caddr_t);
# define generic_quotactl(sb,d,c,t,id,addr)	vfsqops(sb, d, c, t, id, addr)
#else
# define generic_quotactl(sb,d,c,t,id,addr)	(-ENOSYS)
#endif

/*
 * This is the hack letting us get into XFS quota land.
 * It emulates the mechanism which the new quota patches provide.
 */
static int
sbqop(struct super_block *sb, int cmds, short type, qid_t id, caddr_t addr)
{
	int ret = -EINVAL;
	union {
		unsigned int flags;
		struct fs_disk_quota fdq;
		struct fs_quota_stat fqs;
	} u;  /* qualifies valid uses of "addr" */

	switch (cmds) {
		case Q_XQUOTAON:
		case Q_XQUOTAOFF:
		case Q_XQUOTARM:
			if (!sb->s_qop->setxstate)
				break;
			ret = -EFAULT;
			if (copy_from_user(&u.flags, addr, sizeof(u.flags)))
				break;
			ret = sb->s_qop->setxstate(sb, u.flags, cmds);
			break;

		case Q_XGETQSTAT:
			if (sb->s_qop->getxstate)
				ret = sb->s_qop->getxstate(sb, &u.fqs);
			if (!ret && copy_to_user(addr, &u.fqs, sizeof(u.fqs)))
				ret = -EFAULT;
			break;

		case Q_XSETQLIM:
			if (!sb->s_qop->setxquota)
				break;
			ret = -EFAULT;
			if (copy_from_user(&u.fdq, addr, sizeof(u.fdq)))
				break;
			ret = sb->s_qop->setxquota(sb, type, id, &u.fdq);
			break;

		case Q_XGETQUOTA:
			if (sb->s_qop->getxquota)
				ret = sb->s_qop->getxquota(sb, type, id, &u.fdq);
			if (!ret && copy_to_user(addr, &u.fdq, sizeof(u.fdq)))
				ret = -EFAULT;
			break;

		default:
	}
	return ret;
}

asmlinkage long sys_quotactl(int cmd, const char *special, int id, caddr_t addr)
{
	int cmds = 0, type = 0;
	struct super_block *sb = NULL;
	int ret = -EINVAL;
	kdev_t dev;

	lock_kernel();
	cmds = cmd >> SUBCMDSHIFT;
	type = cmd & SUBCMDMASK;

	if ((u_int) type >= MAXQUOTAS)
		goto out;

	ret = -EPERM;
	switch (cmds) {
		case Q_SYNC:
		case Q_GETSTATS:
		case Q_XGETQSTAT:
			break;
		case Q_GETQUOTA:
		case Q_XGETQUOTA:
			if (((type == USRQUOTA && current->euid != id) ||
			     (type == GRPQUOTA && !in_egroup_p(id))) &&
			    !capable(CAP_SYS_ADMIN))
				goto out;
			break;
		default:
			if (!capable(CAP_SYS_ADMIN))
				goto out;
	}

	ret = -EINVAL;
	dev = NODEV;
	if (special != NULL || (cmds != Q_SYNC && cmds != Q_GETSTATS)) {
		mode_t mode;
		struct nameidata nd;

		ret = user_path_walk(special, &nd);
		if (ret)
			goto out;

		dev = nd.dentry->d_inode->i_rdev;
		mode = nd.dentry->d_inode->i_mode;
		path_release(&nd);

		ret = -ENOTBLK;
		if (!S_ISBLK(mode))
			goto out;
		ret = -ENODEV;
		sb = get_super(dev);
		if (!sb)
			goto out;

		if (sb->s_qop) {
			ret = sbqop(sb, cmds, type, id, addr);
			goto out;
		}
	}

	ret = generic_quotactl(sb, dev, cmds, type, id, addr);
out:
	if (sb)
		drop_super(sb);
	unlock_kernel();
	return ret;
}
