#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/pagemap.h>
#include <linux/fs.h>
#include <linux/iobuf.h>
#include <linux/kdb.h>
#include <linux/kdbprivate.h>
#include <linux/blkdev.h>
#include <linux/ctype.h>

MODULE_AUTHOR("SGI");
MODULE_DESCRIPTION("Debug page information");
MODULE_LICENSE("GPL");

#undef	KDB_DO_PAGEBUF
#if	defined(CONFIG_PAGE_BUF) || defined(CONFIG_PAGE_BUF_MODULE)
#define	KDB_DO_PAGEBUF
#define _PAGE_BUF_INTERNAL_ 1
#include <linux/page_buf.h>
#endif	/* CONFIG_PAGE_BUF || CONFIG_PAGE_BUF_MODULE */

/* Standard Linux page stuff */

static char	*pg_flag_vals[] = {
	"PG_locked", "PG_error", "PG_referenced", "PG_uptodate",
	"PG_dirty", "PG_unused_5", "PG_lru", "PG_active",
	"PG_slab", "PG_unused_9", "PG_skip", "PG_highmem",
	"PG_checked", "PG_arch_1", "PG_reserved", "PG_launder",
	NULL };

static char	*bh_state_vals[] = {
	"Uptodate", "Dirty", "Lock", "Req",
	"Mapped", "New", "Async", "Wait_IO",
	"Launder", "JBD",
	/*XFS*/ "Delay",
	NULL };

static char *inode_flag_vals[] = {
	"I_DIRTY_SYNC", "I_DIRTY_DATASYNC", "I_DIRTY_PAGES", "I_LOCK",
	"I_FREEING", "I_CLEAR",
	/*XFS*/ "I_NEW",
	NULL };

static char	*map_flags(unsigned long flags, char *mapping[])
{
	static	char	buffer[256];
	int	index;
	int	offset = 12;

	buffer[0] = '\0';

	for (index = 0; flags && mapping[index]; flags >>= 1, index++) { 
		if (flags & 1) {
			if ((offset + strlen(mapping[index]) + 1) >= 80) {
				strcat(buffer, "\n            ");
				offset = 12;
			} else if (offset > 12) {
				strcat(buffer, " ");
				offset++;
			}
			strcat(buffer, mapping[index]);
			offset += strlen(mapping[index]);
		}
	}

	return (buffer);
}

static char	*page_flags(unsigned long flags)
{
	return(map_flags(flags, pg_flag_vals));
}

static int
kdbm_buffers(int argc, const char **argv, const char **envp,
	struct pt_regs *regs)
{
	struct buffer_head	bh;
	unsigned char *p = (unsigned char *)&bh;
	unsigned long addr;
	long	offset=0;
	int nextarg;
	int diag;
	int i;
	
	if (argc != 1) 
		return KDB_ARGCOUNT;

	nextarg = 1;
	diag = kdbgetaddrarg(argc, argv, &nextarg, &addr, &offset, NULL, regs);
	if (diag) 
		return diag;

	for (i = 0; i < sizeof(struct buffer_head); i++) {
		*p++ = kdba_getword(addr+i, 1);
	}

	kdb_printf("buffer_head at 0x%lx\n", addr);
	kdb_printf("  next 0x%p bno %ld rsec %ld size %d dev 0x%x rdev 0x%x\n",
		bh.b_next, bh.b_blocknr, bh.b_rsector,
		bh.b_size, bh.b_dev, bh.b_rdev);
	kdb_printf("  count %d state 0x%lx [%s] ftime 0x%lx b_list %d b_reqnext 0x%p b_data 0x%p\n",
		bh.b_count.counter, bh.b_state, map_flags(bh.b_state, bh_state_vals),
		bh.b_flushtime, bh.b_list, bh.b_reqnext, bh.b_data);
	kdb_printf("  b_page 0x%p b_this_page 0x%p b_private 0x%p\n",
		bh.b_page, bh.b_this_page, bh.b_private);

	return 0;
}

static int
kdbm_page(int argc, const char **argv, const char **envp,
	struct pt_regs *regs)
{
	struct page	page;
	unsigned char *p = (unsigned char *)&page;
	unsigned long addr;
	long	offset=0;
	int nextarg;
	int diag;
	int i;
	
	if (argc != 1) 
		return KDB_ARGCOUNT;

	nextarg = 1;
	diag = kdbgetaddrarg(argc, argv, &nextarg, &addr, &offset, NULL, regs);
	if (diag) 
		return diag;

	if (addr < PAGE_OFFSET) {
		printk("Treating 0x%lx as page index, page at 0x%p\n",
			addr, &mem_map[addr]);
		addr = (unsigned long) &mem_map[addr];
	}

	for (i = 0; i < sizeof(struct page); i++) {
		*p++ = kdba_getword(addr+i, 1);
	}

	kdb_printf("struct page at 0x%lx\n", addr);
	kdb_printf("  next 0x%p prev 0x%p addr space 0x%p index %lu (offset 0x%x)\n",
		   page.list.next, page.list.prev, page.mapping, page.index,
		   (int)(page.index << PAGE_CACHE_SHIFT));
	kdb_printf("  count %d flags %s virtual 0x%p\n",
		   page.count.counter, page_flags(page.flags),
		   page.virtual);
	kdb_printf("  buffers 0x%p\n", page.buffers);

	return 0;
}

unsigned long
print_request(unsigned long addr)
{
	struct request	rq;
	unsigned char *p = (unsigned char *)&rq;
	int i;

	for (i = 0; i < sizeof(struct request); i++) {
		*p++ = kdba_getword(addr+i, 1);
	}

	kdb_printf("struct request at 0x%lx\n", addr);
	kdb_printf("  rq_dev 0x%x cmd %d errors %d sector %ld nr_sectors %ld\n",
			rq.rq_dev, rq.cmd, rq.errors, rq.sector,
			rq.nr_sectors);

	kdb_printf("  hsect %ld hnrsect %ld nrseg %d nrhwseg %d currnrsect %ld seq %d\n",
			rq.hard_sector, rq.hard_nr_sectors,
			rq.nr_segments, rq.nr_hw_segments,
			rq.current_nr_sectors, rq.elevator_sequence);
	kdb_printf("  ");
#ifdef	STRUCT_REQUEST_HAS_KIOBUF
	kdb_printf("kiobuf 0x%p ", rq.kiobuf);
#endif	/* STRUCT_REQUEST_HAS_KIOBUF */
	kdb_printf("bh 0x%p bhtail 0x%p req_q 0x%p\n\n",
			rq.bh, rq.bhtail, rq.q);

	return (unsigned long) rq.queue.next;
}

#ifdef	KDB_DO_PAGEBUF
static int
kdbm_request(int argc, const char **argv, const char **envp,
	struct pt_regs *regs)
{
	long	offset=0;
	unsigned long addr;
	int nextarg;
	int diag;
	
	if (argc != 1) 
		return KDB_ARGCOUNT;

	nextarg = 1;
	diag = kdbgetaddrarg(argc, argv, &nextarg, &addr, &offset, NULL, regs);
	if (diag) 
		return diag;

	print_request(addr);
	return 0;
}


static int
kdbm_rqueue(int argc, const char **argv, const char **envp,
	struct pt_regs *regs)
{
	struct request_queue	rq;
	unsigned char *p = (unsigned char *)&rq;
	unsigned long addr, head_addr, next;
	long	offset=0;
	int nextarg;
	int diag;
	int i;
	
	if (argc != 1) 
		return KDB_ARGCOUNT;

	nextarg = 1;
	diag = kdbgetaddrarg(argc, argv, &nextarg, &addr, &offset, NULL, regs);
	if (diag) 
		return diag;

	for (i = 0; i < sizeof(struct request_queue); i++) {
		*p++ = kdba_getword(addr+i, 1);
	}

	kdb_printf("struct request_queue at 0x%lx [%s]\n", addr,
			rq.plugged ? "plugged" : "running");
	kdb_printf(" read free_list [0x%p, 0x%p]\n",
			rq.rq[READ].free.prev,
			rq.rq[READ].free.next);
	kdb_printf(" write free_list [0x%p, 0x%p]\n",
			rq.rq[WRITE].free.prev,
			rq.rq[WRITE].free.next);

	i = 0;
	next = (unsigned long)rq.queue_head.next;
	head_addr = addr + offsetof(struct request_queue, queue_head);
	kdb_printf(" request queue: %s\n", next == head_addr ?
		"empty" : "");
	while (next != head_addr) {
		i++;
		next = print_request(next);
	}

	if (i)
		kdb_printf("%d requests found\n", i);

	return 0;
}
#endif	/* KDB_DO_PAGEBUF */


static void
do_buffer(unsigned long addr)
{
	struct buffer_head	bh;
	unsigned char *p = (unsigned char *)&bh;
	int i;
	
	for (i = 0; i < sizeof(struct buffer_head); i++) {
		*p++ = kdba_getword(addr+i, 1);
	}

	kdb_printf("bh 0x%lx bno %8ld [%s]\n", addr, bh.b_blocknr,
		 map_flags(bh.b_state, bh_state_vals));
}

static int
kdbm_inode_pages(int argc, const char **argv, const char **envp,
	struct pt_regs *regs)
{
	struct inode	inode;
	struct address_space ap;
	unsigned char *p = (unsigned char *)&inode;
	unsigned long addr, addr1 = 0;
	long	offset=0;
	int nextarg;
	int diag;
	int i, j;
	int which=0;

	struct list_head *head, *curr;
	
	if (argc < 1) 
		return KDB_ARGCOUNT;

	nextarg = 1;
	diag = kdbgetaddrarg(argc, argv, &nextarg, &addr, &offset, NULL, regs);
	if (diag) 
		return diag;

	if (argc == 2) {
		nextarg = 2;
		diag = kdbgetaddrarg(argc, argv, &nextarg, &addr1,
					&offset, NULL, regs);
		if (diag) 
			return diag;
		kdb_printf("Looking for page index 0x%lx ... \n", addr1);
	}

	for (i = 0; i < sizeof(struct inode); i++) {
		*p++ = kdba_getword(addr+i, 1);
	}

	addr = (unsigned long) inode.i_mapping;
	p = (unsigned char *)&ap;

	for (i = 0; i < sizeof(struct address_space); i++) {
		*p++ = kdba_getword(addr+i, 1);
	}
	
	if (!&inode.i_mapping) goto out;
 again:
	if (which == 0){
	  which=1;
	  head = &inode.i_mapping->clean_pages;
	  kdb_printf("CLEAN  page_struct   index  cnt  flags\n");
	} else if (which == 1) {
	  which=2;
	  head = &inode.i_mapping->dirty_pages;
	  kdb_printf("DIRTY  page_struct   index  cnt  flags\n");
	} else if (which == 2) {
	  which=3;
	  head = &inode.i_mapping->locked_pages;
	  kdb_printf("LOCKED page_struct   index  cnt  flags\n");
	} else {
	  goto out;
	}
	
	if(!head) goto again;
	curr = head->next;
	while (curr != head) {
		struct page 	 page;
		struct list_head curr_struct;

		addr = (unsigned long) list_entry(curr, struct page, list);
		p = (unsigned char *)&page;

		for (j = 0; j < sizeof(struct page); j++)
			*p++ = kdba_getword(addr+j, 1);

		if (!addr1 || page.index == addr1 ||
			(addr1 == -1 && (page.flags & ( 1 << PG_locked))))
		{
			kdb_printf("    0x%lx    %6lu    %5d    0x%lx ",
				addr, page.index, page.count.counter,
				page.flags);
			if (page.buffers)
				do_buffer((unsigned long) page.buffers);
		}

		addr = (unsigned long) curr;
		p = (unsigned char *)&curr_struct;
		for (j = 0; j < sizeof(struct list_head); j++)
			*p++ = kdba_getword(addr+j, 1);

		curr = curr_struct.next;
	}
	goto again;
 out:
	return 0;
}

static int
kdbm_inode(int argc, const char **argv, const char **envp,
	struct pt_regs *regs)
{
	struct inode	inode;
	unsigned char *p = (unsigned char *)&inode;
	unsigned long addr;
	unsigned char *iaddr;
	long	offset=0;
	int nextarg;
	int diag;
	int i;
	
	if (argc != 1) 
		return KDB_ARGCOUNT;

	nextarg = 1;
	diag = kdbgetaddrarg(argc, argv, &nextarg, &addr, &offset, NULL, regs);
	if (diag) 
		return diag;

	for (i = 0; i < sizeof(struct inode); i++) {
		*p++ = kdba_getword(addr+i, 1);
	}

	kdb_printf("struct inode at  0x%lx\n", addr);

	kdb_printf(" i_ino = %lu i_count = %u i_dev = 0x%x i_size %Ld\n",
					inode.i_ino, atomic_read(&inode.i_count),
					inode.i_dev, inode.i_size);

	kdb_printf(" i_mode = 0x%x  i_nlink = %d  i_rdev = 0x%x\n",
					inode.i_mode, inode.i_nlink,
					inode.i_rdev);

	kdb_printf(" i_hash.nxt = 0x%p i_hash.prv = 0x%p\n",
					inode.i_hash.next, inode.i_hash.prev);

	kdb_printf(" i_list.nxt = 0x%p i_list.prv = 0x%p\n",
					inode.i_list.next, inode.i_list.prev);

	kdb_printf(" i_dentry.nxt = 0x%p i_dentry.prv = 0x%p\n",
					inode.i_dentry.next,
					inode.i_dentry.prev);

	kdb_printf(" i_dirty_buffers.nxt = 0x%p i_dirty_buffers.prv = 0x%p\n",
					inode.i_dirty_buffers.next,
					inode.i_dirty_buffers.prev);

	kdb_printf(" i_sb = 0x%p i_op = 0x%p i_data = 0x%lx nrpages = %lu\n",
					inode.i_sb, inode.i_op,
					addr + offsetof(struct inode, i_data),
					inode.i_data.nrpages);
	kdb_printf(" i_mapping = 0x%p\n i_flags 0x%x i_state 0x%lx [%s]",
			   inode.i_mapping, inode.i_flags,
			   inode.i_state,
			   map_flags(inode.i_state, inode_flag_vals));
	
	iaddr  = (char *)addr;
	iaddr += offsetof(struct inode, u);

	kdb_printf("  fs specific info @ 0x%p\n", iaddr);

	return (0);
}

static int
kdbm_kiobuf(int argc, const char **argv, const char **envp,
	struct pt_regs *regs)
{
	struct kiobuf kiobuf;
	struct page	page;
	unsigned char *p = (unsigned char *)&kiobuf;
	struct page	*page_array[64];
	unsigned long addr;
	long	offset=0;
	int nextarg;
	int diag;
	int i, j;
	
	if (argc != 1) 
		return KDB_ARGCOUNT;

	nextarg = 1;
	diag = kdbgetaddrarg(argc, argv, &nextarg, &addr, &offset, NULL, regs);
	if (diag) 
		return diag;

	for (i=0; i<sizeof(struct kiobuf); i++) {
		*p++ = kdba_getword(addr+i, 1);
	}

	kdb_printf("kiobuf at 0x%lx\n", addr);
	kdb_printf("  nr_pages %d array_len %d offset 0x%x length 0x%x\n",
		   kiobuf.nr_pages, kiobuf.array_len,
		   kiobuf.offset, kiobuf.length);
	if (kiobuf.maplist != (((struct kiobuf *)addr)->map_array)) {
		p = (char *)page_array;
		for (i=0; i < (kiobuf.nr_pages * sizeof(struct page *)); i++) {
			*p++ = kdba_getword((kdb_machreg_t)kiobuf.maplist + i, 1);
		}
		kiobuf.maplist = page_array;
	}
	kdb_printf("  errno %d", kiobuf.errno);
#ifdef	KDB_DO_PAGEBUF
#ifdef	KIOBUF_IO
	kdb_printf("  pb 0x%p", (page_buf_t *)kiobuf.k_dev_id);
#endif
#endif	/* KDB_DO_PAGEBUF */
	kdb_printf("\n");
	kdb_printf("    page_struct   page_addr     cnt  flags\n");
	for (i = 0; i < kiobuf.nr_pages; i++) {
		addr = (unsigned long) kiobuf.maplist[i];
		p = (unsigned char *)&page;

		for (j = 0; j < sizeof(struct page); j++) {
			*p++ = kdba_getword(addr+j, 1);
		}
		kdb_printf("    0x%p    0x%p    %d    0x%lx\n",
			   kiobuf.maplist[i], page.virtual,
			   page.count.counter, page.flags);
	}

	return (0);
}

#ifdef	KDB_DO_PAGEBUF

/* pagebuf stuff */

static char	*pb_flag_vals[] = {
	"READ", "WRITE", "MAPPED", "PARTIAL",
	"ASYNC", "NONE", "DELWRI", "FREED", "SYNC",
	"MAPPABLE", "FS_RESERVED_1", "FS_RESERVED_2", "RELEASE",
	"LOCK", "TRYLOCK", "ALLOCATE", "FILE_ALLOCATE", "DONT_BLOCK",
	"DIRECT", "LOCKABLE", "NEXT_KEY", "ENTER_PAGES",
	"ALL_PAGES_MAPPED", "SOME_INVALID_PAGES", "ADDR_ALLOCATED",
	"MEM_ALLOCATED", "GRIO", "FORCEIO", "SHUTDOWN",
	NULL };

static char	*pbm_flag_vals[] = {
	"EOF", "HOLE", "DELAY", "FLUSH_OVERLAPS",
	"READAHEAD", "UNWRITTEN", "DONTALLOC", "NEW",
	NULL };

static char	*pb_flags(page_buf_flags_t pb_flag)
{
	return(map_flags((unsigned long) pb_flag, pb_flag_vals));
}

static int
kdbm_pb_flags(int argc, const char **argv, const char **envp, struct pt_regs *regs)
{
	unsigned long flags;
	int diag;
	
	if (argc != 1) 
		return KDB_ARGCOUNT;

	diag = kdbgetularg(argv[1], &flags);
	if (diag) 
		return diag;

	kdb_printf("pb flags 0x%lx = %s\n", flags, pb_flags(flags));

	return 0;
}

static int
kdbm_pb(int argc, const char **argv, const char **envp, struct pt_regs *regs)
{
	page_buf_private_t bp;
	unsigned char *p = (unsigned char *)&bp;
	unsigned long addr;
	long	offset=0;
	int nextarg;
	int diag;
	int i;
	
	if (argc != 1) 
		return KDB_ARGCOUNT;

	nextarg = 1;
	diag = kdbgetaddrarg(argc, argv, &nextarg, &addr, &offset, NULL, regs);
	if (diag) 
		return diag;

	for (i=0; i<sizeof(page_buf_private_t); i++) {
		*p++ = kdba_getword(addr+i, 1);
	}

	kdb_printf("page_buf_t at 0x%lx\n", addr);
	kdb_printf("  pb_flags %s\n", pb_flags(bp.pb_common.pb_flags));
	kdb_printf("  pb_target 0x%p pb_hold %d pb_next 0x%p pb_prev 0x%p\n",
		   bp.pb_common.pb_target, bp.pb_common.pb_hold,
		   bp.pb_common.pb_list.next, bp.pb_common.pb_list.prev);
	kdb_printf("  pb_file_offset 0x%Lx pb_buffer_length 0x%x pb_addr 0x%p\n",
		   bp.pb_common.pb_file_offset, bp.pb_common.pb_buffer_length,
		   bp.pb_common.pb_addr);
	kdb_printf("  pb_bn 0x%Lx pb_count_desired 0x%x\n",
		   bp.pb_common.pb_bn,
		   bp.pb_common.pb_count_desired);
	kdb_printf("  pb_io_remaining %d pb_error %d\n",
		   bp.pb_io_remaining.counter, bp.pb_common.pb_error);
	kdb_printf("  pb_page_count %d pb_offset 0x%x pb_pages 0x%p\n",
		bp.pb_common.pb_page_count, bp.pb_common.pb_offset,
		bp.pb_common.pb_pages);
#ifdef PAGEBUF_LOCK_TRACKING
	kdb_printf("  pb_iodonesema (%d,%d) pb_sema (%d,%d) pincount (%d) last holder 0x%p\n",
		   bp.pb_iodonesema.count.counter, bp.pb_iodonesema.sleepers,
		   bp.pb_sema.count.counter, bp.pb_sema.sleepers,
		   bp.pb_pin_count.counter, bp.pb_last_holder);
#else
	kdb_printf("  pb_iodonesema (%d,%d) pb_sema (%d,%d) pincount (%d)\n",
		   bp.pb_iodonesema.count.counter, bp.pb_iodonesema.sleepers,
		   bp.pb_sema.count.counter, bp.pb_sema.sleepers,
		   bp.pb_pin_count.counter);
#endif
	if (bp.pb_common.pb_fspriv || bp.pb_common.pb_fspriv2) {
		kdb_printf(  "pb_fspriv 0x%p pb_fspriv2 0x%p\n",
			   bp.pb_common.pb_fspriv, bp.pb_common.pb_fspriv2);
	}

	return 0;
}

/* XXXXXXXXXXXXXXXXXXXXXX */
/* The start of this deliberately looks like a read_descriptor_t in layout */
typedef struct {
	read_descriptor_t io_rdesc;

	/* 0x10 */
	page_buf_rw_t io_dir;	/* read or write */
	loff_t io_offset;	/* Starting offset of I/O */
	int io_iovec_nr;	/* Number of entries in iovec */

	/* 0x20 */
	struct iovec **io_iovec;	/* iovec list indexed by iovec_index */
	loff_t io_iovec_offset;	/* offset into current iovec. */
	int io_iovec_index;	/* current iovec being processed */
	unsigned int io_sshift;	/* sector bit shift */
	loff_t io_i_size;	/* size of the file */
} pb_io_desc_t;

static int
kdbm_pbiodesc(int argc, const char **argv, const char **envp,
	struct pt_regs *regs)
{
	pb_io_desc_t	pbio;
	unsigned char *p = (unsigned char *)&pbio;
	unsigned long addr;
	long	offset=0;
	int nextarg;
	int diag;
	int i;
	
	if (argc != 1) 
		return KDB_ARGCOUNT;

	nextarg = 1;
	diag = kdbgetaddrarg(argc, argv, &nextarg, &addr, &offset, NULL, regs);
	if (diag) 
		return diag;

	for (i = 0; i < sizeof(pb_io_desc_t); i++) {
		*p++ = kdba_getword(addr+i, 1);
	}

	kdb_printf("pb_io_desc_t at 0x%lx\n", addr);
	kdb_printf("  io_rdesc [ written 0x%x count 0x%x buf 0x%p error %d ]\n",
			pbio.io_rdesc.written, pbio.io_rdesc.count, 
			pbio.io_rdesc.buf, pbio.io_rdesc.error); 

	kdb_printf("  io_dir %d io_offset 0x%Lx io_iovec_nr 0x%d\n",
			pbio.io_dir, pbio.io_offset, pbio.io_iovec_nr);

	kdb_printf("  io_iovec 0x%p io_iovec_offset 0x%Lx io_iovec_index 0x%d\n",
		pbio.io_iovec, pbio.io_iovec_offset, pbio.io_iovec_index);

	kdb_printf("  io_sshift 0x%d io_i_size 0x%Lx\n",
		pbio.io_sshift, pbio.io_i_size);

	return 0;
}

static int
kdbm_pbmap(int argc, const char **argv, const char **envp,
	struct pt_regs *regs)
{
	page_buf_bmap_t	pbm;
	unsigned char *p = (unsigned char *)&pbm;
	unsigned long addr;
	long	offset=0;
	int nextarg;
	int diag;
	int i;
	
	if (argc != 1) 
		return KDB_ARGCOUNT;

	nextarg = 1;
	diag = kdbgetaddrarg(argc, argv, &nextarg, &addr, &offset, NULL, regs);
	if (diag) 
		return diag;

	for (i = 0; i < sizeof(page_buf_bmap_t); i++) {
		*p++ = kdba_getword(addr+i, 1);
	}

	kdb_printf("page_buf_bmap_t at 0x%lx\n", addr);
	kdb_printf("  pbm_bn 0x%Lx pbm_offset 0x%Lx pbm_delta 0x%x pbm_bsize 0x%x\n",
		pbm.pbm_bn, pbm.pbm_offset, pbm.pbm_delta, pbm.pbm_bsize);

	kdb_printf("  pbm_flags %s\n", map_flags(pbm.pbm_flags, pbm_flag_vals));

	return 0;
}

#ifdef	PAGEBUF_TRACE
#include <linux/page_buf_trace.h>

#define EV_SIZE	(sizeof(event_names)/sizeof(char *))

void
pb_trace_core(
	unsigned long match,
	char	*event_match,
	unsigned long long offset,
	long long mask)
{
	extern struct pagebuf_trace_buf pb_trace;
        int i, total, end;
	pagebuf_trace_t	*trace;
	char	*event;
	char	value[10];

	end = pb_trace.start - 1;
	if (end < 0)
		end = PB_TRACE_BUFSIZE - 1;

	if (match && (match < PB_TRACE_BUFSIZE)) {
		for (i = pb_trace.start, total = 0; i != end; i = CIRC_INC(i)) {
			trace = &pb_trace.buf[i];
			if (trace->pb == 0) 
				continue;
			total++;
		}
		total = total - match;
		for (i = pb_trace.start; i != end && total; i = CIRC_INC(i)) {
			trace = &pb_trace.buf[i];
			if (trace->pb == 0) 
				continue;
			total--;
		}
		match = 0;
	} else
		i = pb_trace.start;
	for ( ; i != end; i = CIRC_INC(i)) {
		trace = &pb_trace.buf[i];

		if (offset) {
			if ((trace->offset & ~mask) != offset)
				continue;
		}

		if (trace->pb == 0) 
			continue;

		if ((match != 0) && (trace->pb != match))
			continue;

		if ((trace->event < EV_SIZE) && event_names[trace->event]) {
			event = event_names[trace->event];
		} else if (trace->event == 1000) {
			event = (char *)trace->misc;
		} else {
			event = value;
			sprintf(value, "%8d", trace->event);
		}

		if (event_match && strcmp(event, event_match)) {
			continue;
		}


		kdb_printf("pb 0x%lx [%s] (hold %u lock %d) misc 0x%p",
			   trace->pb, event,
			   trace->hold, trace->lock_value,
			   trace->misc);
		kdb_symbol_print((unsigned int)trace->ra, NULL,
			KDB_SP_SPACEB|KDB_SP_PAREN|KDB_SP_NEWLINE);
		kdb_printf("    offset 0x%Lx size 0x%x task 0x%p\n",
			   trace->offset, trace->size, trace->task);
		kdb_printf("    flags: %s\n",
			   pb_flags(trace->flags));
	}
}


static int
kdbm_pbtrace_offset(int argc, const char **argv, const char **envp,
	struct pt_regs *regs)
{
        long mask = 0;
	unsigned long offset = 0;
        int diag;

        if (argc > 2)
                return KDB_ARGCOUNT;

	if (argc > 0) {
		diag = kdbgetularg(argv[1], &offset);
		if (diag)
			return diag;
	}

	if (argc > 1) {
		diag = kdbgetularg(argv[1], &mask);
		if (diag)
			return diag;
	}

	pb_trace_core(0, NULL, (unsigned long long)offset,
			       (long long)mask);	/* sign extent mask */
	return 0;
}

static int
kdbm_pbtrace(int argc, const char **argv, const char **envp,
	struct pt_regs *regs)
{
        unsigned long addr = 0;
        int diag, nextarg;
	long offset = 0;
	char	*event_match = NULL;

        if (argc > 1)
                return KDB_ARGCOUNT;

	if (argc == 1) {
		if (isupper(argv[1][0]) || islower(argv[1][0])) {
			event_match = (char *)argv[1];
			printk("event match on \"%s\"\n", event_match);
			argc = 0;
		} else {
			diag = kdbgetaddrarg(argc, argv, &nextarg, &addr, &offset, NULL, regs);
			if (diag) {
				printk("failed to parse %s as a number\n",
								argv[1]);
				return diag;
			}
		}
	}

	pb_trace_core(addr, event_match, 0LL, 0LL);
	return 0;
}

#else	/* PAGEBUF_TRACE */
static int
kdbm_pbtrace(int argc, const char **argv, const char **envp,
	struct pt_regs *regs)
{
	kdb_printf("pagebuf tracing not compiled in\n");

	return 0;
}
#endif	/* PAGEBUF_TRACE */
#endif	/* KDB_DO_PAGEBUF */

static int __init kdbm_pg_init(void)
{
	kdb_register("kiobuf", kdbm_kiobuf, "<vaddr>", "Display kiobuf", 0);
	kdb_register("page", kdbm_page, "<vaddr>", "Display page", 0);
	kdb_register("inode", kdbm_inode, "<vaddr>", "Display inode", 0);
	kdb_register("bh", kdbm_buffers, "<buffer head address>", "Display buffer", 0);
	kdb_register("inode_pages", kdbm_inode_pages, "<inode *>", "Display pages in an inode", 0);
#ifdef	KDB_DO_PAGEBUF
	kdb_register("pb", kdbm_pb, "<vaddr>", "Display page_buf_t", 0);
	kdb_register("pbflags", kdbm_pb_flags, "<flags>", "Display page buf flags", 0);
	kdb_register("pbiodesc", kdbm_pbiodesc, "<pb_io_desc_t *>", "Display I/O Descriptor", 0);
	kdb_register("pbmap", kdbm_pbmap, "<page_buf_bmap_t *>", "Display Bmap", 0);
	kdb_register("pbtrace", kdbm_pbtrace, "<vaddr>|<count>", "page_buf_t trace", 0);
#ifdef PAGEBUF_TRACE
	kdb_register("pboffset", kdbm_pbtrace_offset, "<addr> [<mask>]", "page_buf_t trace", 0);
#endif
	kdb_register("req", kdbm_request, "<vaddr>", "dump request struct", 0);
	kdb_register("rqueue", kdbm_rqueue, "<vaddr>", "dump request queue", 0);

#endif	/* KDB_DO_PAGEBUF */
	
	return 0;
}


static void __exit kdbm_pg_exit(void)
{
	kdb_unregister("kiobuf");
	kdb_unregister("page");
	kdb_unregister("inode");
	kdb_unregister("bh");
	kdb_unregister("inode_pages");
	kdb_unregister("pb");
	kdb_unregister("pbflags");
	kdb_unregister("pbmap");
	kdb_unregister("pbiodesc");
	kdb_unregister("pbtrace");
#ifdef PAGEBUF_TRACE
	kdb_unregister("pboffset");
#endif
	kdb_unregister("req");
	kdb_unregister("rqueue");
}

module_init(kdbm_pg_init) 
module_exit(kdbm_pg_exit)
